export VARNAME_CI_SECRET="EROTEMIC_CI_SECRET"
export VARNAME_TWINE_USERNAME="TWINE_USERNAME"
export VARNAME_TWINE_PASSWORD="TWINE_PASSWORD"
export VARNAME_TEST_TWINE_USERNAME="TEST_TWINE_USERNAME"
export VARNAME_TEST_TWINE_PASSWORD="TEST_TWINE_PASSWORD"
export GPG_IDENTIFIER="=Erotemic-CI <erotemic@gmail.com>"
